__all__ = [
    'all',
    'assemble',
    'bam2reads',
    'clean',
    'fixstart',
    'get_dnaa',
    'mapreads',
    'merge',
    'minimus2',
    'progcheck',
    'version',
]

from circlator.tasks import *
